#define _BuildMode_Lib 0
#define _BuildMode_Dll 1
#define _BuildMode_App 2



#define _Export _declspec (dllexport)
#define _Import _declspec (dllimport)



#if _BuildMode == _BuildMode_Lib

  __pragma (message ("Static library build mode"))

  #define _LinkageDecl _Export
  #define _LinkageDef _Export

#elif _BuildMode == _BuildMode_Dll
  
  __pragma (message ("DLL build mode"))

  #define _LinkageDecl
  #define _LinkageDef

#elif _BuildMode == _BuildMode_App
  
  __pragma (message ("Application build mode"))

  #define _LinkageDecl
  #define _LinkageDef

#else

  #error Incorrect _BuildMode value

#endif _BuildMode





namespace ThinAudioGateway {





// #############################################################################################
//
// Classes
// =============================================================================================
// 
// #############################################################################################





typedef UINT LocSize;





__if_not_exists (KSDATAFORMAT_WAVEFORMATEXTENSIBLE) {

  struct KSDATAFORMAT_WAVEFORMATEXTENSIBLE {

    KSDATAFORMAT DataFormat;
    WAVEFORMATEXTENSIBLE WaveFormatExt;

  };

}





#if _Debug

class _LinkageDef DebugContext {

  wchar_t Name [32];
  UINT PrintLevel;

  DebugContext const * Parent;
  DebugContext const * Root;

public:

  DebugContext (DebugContext const * xParent, UINT const xPrintLevel, PCTSTR const _Fmt, ...);

  void SetParent (DebugContext const * const xParent);

  void SetName (PCTSTR const _Fmt, va_list const Args);
  void _cdecl SetName (PCTSTR const _Fmt, ...);

  void SetPrintLevel (UINT const Level);

  void Printf (UINT const Level, PCTSTR const _Fmt, va_list const Args) const;
  void _cdecl Printf (UINT const Level, PCTSTR const _Fmt, ...) const;

};

#endif





template <UINT BufSize> class TextInPlace {

  public:

  // Buffer is defined public to allow derived classes to obtain its size via sizeof ().

  wchar_t Buf [BufSize];

  operator wchar_t const * () const { return Buf; }
  operator wchar_t * () { return Buf; }

};





class _LinkageDef ErrorText_i : public TextInPlace <256> {

  public:

  ErrorText_i (DWORD const Code) { ToStr (Buf, NumArrayElems (Buf), Code); }

  // HRESULT is LONG too

  ErrorText_i (LONG const Code) { ToStr (Buf, NumArrayElems (Buf), static_cast <DWORD> (Code)); }

  static PCTSTR ToStr (PTSTR const Buf, UINT const MaxLen, DWORD const Code);

};

// XxxText () functions must be macros to properly handle temporary objects

#define ErrorText(Code) (static_cast <PCTSTR> (ErrorText_i (Code)))
#define LastErrorText() (static_cast <PCTSTR> (ErrorText_i (GetLastError ())))





class GuidText_i : public TextInPlace <40> {

  public:

  GuidText_i (GUID const & Guid) { ToStr (Buf, NumArrayElems (Buf), Guid); }

  static PTSTR ToStr (PTSTR const Buffer, UINT const MaxLen, GUID const & Guid);

};

// XxxText () functions must be macros to properly handle temporary objects

#define GuidText(Guid) (static_cast <PCTSTR> (GuidText_i (Guid)))





class _LinkageDef WaveFormat {

public:

  DataFormatDesc Fmt;

  WaveFormat () {
    Fmt.SamplingRate = 0;
    Fmt.BitsPerSample = 0;
    Fmt.NumChannels = 0;
    Fmt.BytesPerContainer = 0;
    Fmt.ChannelConfig = 0;
  }

  WaveFormat (
    UINT const SR, UINT const BPS, UINT const NC, UINT const BPC = 0, UINT const CC = UINT (-1)
  ) {
    Init (SR, BPS, NC, BPC, CC);
  }

  HRESULT Init (
    UINT const SR, UINT const BPS, UINT const NC, UINT const BPC = 0, UINT const CC = UINT (-1)
  );

  DataFormatDesc const & GetFormat () const { return Fmt; }
  DataFormatDesc & GetFormat () { return Fmt; }

  UINT GetSR () const { return Fmt.SamplingRate; }
  UINT GetBPS () const { return Fmt.BitsPerSample; }
  UINT GetNC () const { return Fmt.NumChannels; }
  UINT GetBPC () const { return Fmt.BytesPerContainer; }
  UINT GetCC () const { return Fmt.ChannelConfig; }

  UINT GetFrameSize () const { return Fmt.BytesPerContainer * Fmt.NumChannels; }

  static UINT GetStdChannelMask (UINT const NumChans);

  UINT BytesToFrames (UINT const Bytes) const { return Bytes / GetFrameSize (); }
  UINT FramesToBytes (UINT const Frames) const { return Frames * GetFrameSize (); }
  UINT MsToBytes (UINT const Ms) const { return FramesToBytes (Ms * Fmt.SamplingRate / 1000); }

  HRESULT FromWfx (WAVEFORMATEX const & Wfx);
  void ToWfx (WAVEFORMATEXTENSIBLE & Wfx) const;
  void ToKdf (KSDATAFORMAT_WAVEFORMATEXTENSIBLE & Kdf) const;
  PCTSTR ToStr (PTSTR const Buffer, UINT const MaxLen) const;

};



class WfxText_i : public TextInPlace <50> {

  public:

  WfxText_i (WAVEFORMATEXTENSIBLE const & Wfx) {

    WaveFormat Fmt;

    Fmt.FromWfx (Wfx.Format);
    Fmt.ToStr (Buf, NumArrayElems (Buf));
    
  }

  WfxText_i (WAVEFORMATEX const & Wfx) {

    WaveFormat Fmt;

    Fmt.FromWfx (Wfx);
    Fmt.ToStr (Buf, NumArrayElems (Buf));
    
  }

  WfxText_i (WaveFormat const & Fmt) {

    Fmt.ToStr (Buf, NumArrayElems (Buf));
    
  }

};

// XxxText () functions must be macros to properly handle temporary objects

#define WfxText(Fmt) (static_cast <PCTSTR> (WfxText_i (Fmt)))





class CallbackInterface {

public:

  virtual LocSize CopyData (void * const Addr, LocSize const Bytes) = 0;

};





class TagPipe;





class _LinkageDef TagDriver {

  FakeAssignmentOperator (TagDriver)
  FakeCopyConstructor (TagDriver)

  GUID const ClassId;
  wchar_t InterfacePath [MAX_PATH];

#if _Debug

  DebugContext Dbg;

#endif // _Debug

  HRESULT FindInterface ();

public:

  TagDriver (GUID const & xClassId);
  ~TagDriver ();

  bool IsOpen () const { return (*InterfacePath != L'\0'); }

  GUID const & GetClassId () const { return ClassId; }

  static HRESULT _stdcall Create (TagDriver * & Obj, GUID const & xClassId);
  void Delete ();

  HRESULT Open ();
  void Close ();

  HRESULT CreatePipe (TagPipe * & Pipe, bool const Capture);

  void Check () const {

#if _Debug

#endif

  }
  
};





class _LinkageDef TagPipe {

  FakeAssignmentOperator (TagPipe)
  FakeCopyConstructor (TagPipe)

  TagDriver const * Drv;        // Driver reference
  bool const Capture;           // Capture flag

  WaveFormat Fmt;               // Wave format
  UINT FrameSize;               // Frame size in bytes

  HANDLE Connection;            // Connection handle
  CommonDataDesc * CommonData;  // Driver/host common data block address

  HRESULT GetCommonDataAddress (CommonDataDesc * & CD) const;

#if _Debug

  DebugContext Dbg;

#endif // _Debug

public:

  TagPipe (TagDriver const * xDrv, bool const Capt);
  ~TagPipe ();

  bool IsCapt () const { return Capture; }
  bool IsOpen () const { return (Connection != 0); }
  PCTSTR GetType () const { return IsCapt ()? L"capture" : L"render"; }

  bool IsRunningState () const {
    return (
      IsOpen () && CommonData != nullptr
      && CommonData->Drv.FI.HasStream
      && CommonData->Drv.SI.BufferFrames != 0
      && CommonData->Drv.SI.State == KSSTATE_RUN
    );
  }

  void Delete ();

  HRESULT Open (PCTSTR const DriverInterface);
  void Close ();

  CommonDataDesc const & GetCommonData () const { Assert (CommonData != nullptr); return *CommonData; }
  CommonDataDesc & GetCommonData () { Assert (CommonData != nullptr); return *CommonData; }

  HRESULT DriverRequest (
    UINT const Func,
    DriverReq & Req, LocSize const ReqSize,
    void * const Data = nullptr, LocSize const DataSize = 0,
    LocSize * const RetSize = nullptr
  ) const;

  HRESULT SetFormat (
    UINT const SR, UINT const BPS, UINT const NC, UINT const BPC = 0, UINT const CC = UINT (-1)
  );

  HRESULT SetMinimalBufferSize (UINT const Frames);
  HRESULT SetNotificationEvent (HANDLE const Event);

  HRESULT Notify (NTSTATUS const Code);

  HRESULT AdvanceDevicePosition (UINT const Frames);
  HRESULT ResetDevicePosition ();

  void Check () const {

#if _Debug

    Assert (IsOpen () == (CommonData != nullptr));

#endif

  }
  
};





class _LinkageDef TagHost {

  FakeCopyConstructor (TagHost)
  FakeAssignmentOperator (TagHost)

  TagDriver & Driver;
  bool const Capture;
  CallbackInterface & Callback;

  HANDLE Event;
  TagPipe * Pipe;
  WaveFormat Fmt;
  bool Started;

  bool volatile EventLoopActive;
  bool volatile ExitRequest;

#if _Debug

  DebugContext Dbg;

#endif // _Debug

public:

  TagHost (TagDriver & Drv, bool const Capt, CallbackInterface & CB);
  ~TagHost ();

  bool IsCapt () const { return Capture; }
  bool IsStarted () const { return Started; }
  PCTSTR GetType () const { return IsCapt ()? L"capture" : L"render"; }
  CommonDataDesc const & GetCommonData () const;

  static HRESULT _stdcall Create (TagHost * & Obj, TagDriver & Drv, bool const Capture, CallbackInterface & CB);
  void Delete ();

  HRESULT Start (WaveFormat const & Format, UINT const MinBufMs);
  HRESULT Stop ();

  HRESULT EventLoop (UINT const PollPeriodMs);

  void Check () const {

#if _Debug

    Assert (!IsStarted () || Pipe != nullptr && Event != 0);

#endif

  }
  
};





// #############################################################################################
//
// Prorotypes
// =============================================================================================
// 
// #############################################################################################





void _LinkageDecl _cdecl Error (PCTSTR const Format, ...);





} // namespace
