#include "stdafx.h"





namespace ThinAudioGateway {





// #############################################################################################
//
// Variables
// =============================================================================================
//
// #############################################################################################





// #############################################################################################
//
// Functions
// =============================================================================================
//
// #############################################################################################





#if _Debug





DebugContext::DebugContext (DebugContext const * xParent, UINT const xPrintLevel, PCTSTR const _Fmt, ...) {

  SetParent (xParent);
  SetPrintLevel (xPrintLevel);

  Assert (_Fmt != nullptr);

  va_list Args;

  va_start (Args, _Fmt);
  SetName (_Fmt, Args);
  va_end (Args);

}





void DebugContext::SetParent (DebugContext const * const xParent) {

  Parent = xParent;

  if (xParent != nullptr) {

    if (xParent->Parent == nullptr) {

      Root = Parent;

    }

  }

}





void DebugContext::SetName (PCTSTR const _Fmt, va_list const Args) {

  Assert (_Fmt != nullptr && Args != nullptr);

  if (Parent != nullptr && Parent->Parent != nullptr) {

    wcscpy (Name, Parent->Name);
    wcscat (Name, L"/");

  }

  wvsprintf (Name, _Fmt, Args);

}





void _cdecl DebugContext::SetName (PCTSTR const _Fmt, ...) {

  Assert (_Fmt != nullptr);

  va_list Args;

  va_start (Args, _Fmt);
  SetName (_Fmt, Args);
  va_end (Args);

}





void DebugContext::SetPrintLevel (UINT const Level) {

  PrintLevel = Level;

}





void DebugContext::Printf (UINT const Level, PCTSTR const _Fmt, va_list const Args) const {

  Assert (Args != nullptr);

  if (Level >= PrintLevel) {

    wchar_t Buf [300];

    wchar_t const Delim = (Level == 9)? L'*' : L' ';

    PCTSTR const RootName = (Root != nullptr)? Root->Name : Name;

    int N = wsprintf (
      Buf,
      L"%s%c%06u%c%4u%c",
      RootName, Delim, GetTickCount () % 100000, Delim, GetCurrentThreadId (), Delim
    );

    if (Parent != nullptr) {

      wcscat (Buf, Name);
      wcscat (Buf, L": ");
      N += static_cast <int> (wcslen (Name)) + 2;

    }

    N += wvsprintf (Buf + N, _Fmt, Args);

    wcscpy (Buf + N, L"\n");

    Assert (wcslen (Buf) < NumArrayElems (Buf));

    OutputDebugString (Buf);

  }

}





void DebugContext::Printf (UINT const Level, PCTSTR const _Fmt, ...) const {

  if (Level >= PrintLevel) {

    va_list Args;

    va_start (Args, _Fmt);
    Printf (Level, _Fmt, Args);
    va_end (Args);

  }

}





#endif // _Debug





PCTSTR ErrorText_i::ToStr (PTSTR const Buf, UINT const MaxLen, DWORD const Code) {

  DWORD const LastErr = GetLastError ();

  PTSTR cp = Buf;
  UINT Avail = MaxLen;

  _snwprintf (
    cp, Avail,
    L"%s%u/%u - ",
    (Code & APPLICATION_ERROR_MASK)? L"u" : L"",
    static_cast <UINT> (HRESULT_FACILITY (Code)), static_cast <UINT> (HRESULT_CODE (Code))
  );

  cp [Avail - 1] = 0;

  UINT const Offset = static_cast <UINT> (wcslen (Buf));

  cp += Offset;
  Avail -= Offset;

  UINT Len;

  Len = FormatMessage (
    FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS | FORMAT_MESSAGE_MAX_WIDTH_MASK,
    nullptr,
    Code,
    MAKELANGID (LANG_NEUTRAL, SUBLANG_DEFAULT),
    cp,
    Avail,
    nullptr
  );

  if (Len) {

    cp += Len - 1;

    if (*cp == L' ') {

      *cp = L'\0';

    }

  } else if (cp > Buf + 3 && cp [-1] == L' ' && cp [-2] == L'-' && cp [-3] == L' ') {

    cp [-3] = L'\0';             // Cut " - "

  }

  Assert (wcslen (Buf) < MaxLen);

  SetLastError (LastErr);

  return Buf;

}





PTSTR GuidText_i::ToStr (PTSTR const Buffer, UINT const MaxLen, GUID const & Guid) {

  wchar_t Str [40];

  if (&Guid) {

    _snwprintf (
      Str, NumArrayElems (Str),
      L"%08x-%04x-%04x-%02x%02x-%02x%02x%02x%02x%02x%02x",
      Guid.Data1,
      Guid.Data2, Guid.Data3,
      Guid.Data4 [0], Guid.Data4 [1],
      Guid.Data4 [2], Guid.Data4 [3], Guid.Data4 [4],
      Guid.Data4 [5], Guid.Data4 [6], Guid.Data4 [7]
    );

  } else {

    wcscpy (Str, L"GUID_NULL");

  }

  UINT const Len = static_cast <UINT> (wcslen (Str));
  UINT const ActualLen = min (Len, MaxLen - 1);

  memcpy (Buffer, Str, StrLenToSz (ActualLen));

  Buffer [ActualLen] = L'\0';

  return Buffer;

}





//
// Common error message function
// =============================================================================================
//

void Error (PCTSTR const Format, va_list const Args) {

  DWORD const LastErr = GetLastError ();

  wchar_t Text [512];

  _vsnwprintf (Text, NumArrayElems (Text), Format, Args);

  Debug (9, L"Error: %.128s", Text);

  MessageBox (HWND_DESKTOP, Text, L"Error", MB_ICONERROR | MB_OK);

  SetLastError (LastErr);

}





//
// Error message with no help context
// =============================================================================================
//

void _cdecl Error (PCTSTR const Format, ...) {

  va_list Args;

  va_start (Args, Format);
  Error (Format, Args);
  va_end (Args);

}





HRESULT WaveFormat::Init (UINT const SR, UINT const BPS, UINT const NC, UINT const BPC, UINT const CC) {

  HRESULT Res = E_I_PROCESSING_NOT_FINISHED;

  do {

    Debug (
      7,
      L"Initializing format with SR %u, BPS %u, NC %u, BPC %u, CC %#x",
      SR, BPS, NC, BPC, CC
    );

    if (!IsInRange (SR, SamplingRateMin, SamplingRateMax)) {

      Res = HRESULT_FROM_WIN32 (ERROR_BAD_FORMAT);

      Debug (9, L"SR (%u) is out of %u..%u", SR, SamplingRateMin, SamplingRateMax);

      break;

    }

    if (!IsInRange (BPS, BitsPerSampleMin, BitsPerSampleMax)) {

      Res = HRESULT_FROM_WIN32 (ERROR_BAD_FORMAT);

      Debug (9, L"BPS (%u) is out of %u..%u", BPS, BitsPerSampleMin, BitsPerSampleMax);

      break;

    }

    if (!IsInRange (NC, NumChannelsMin, NumChannelsMax)) {

      Res = HRESULT_FROM_WIN32 (ERROR_BAD_FORMAT);

      Debug (9, L"NC (%u) is out of %u..%u", NC, NumChannelsMin, NumChannelsMax);

      break;

    }

    if (BPC > 4) {

      Res = HRESULT_FROM_WIN32 (ERROR_BAD_FORMAT);

      Debug (9, L"BPC (%u) is out of 0..4", BPC);

      break;

    }

    if (BPC != 0 && BPC < (BPS + 7) / 8) {

      Res = HRESULT_FROM_WIN32 (ERROR_BAD_FORMAT);

      Debug (9, L"BPC (%u) cannot hold %u bits", BPC, BPS);

      break;

    }

    Fmt.SamplingRate = SR;
    Fmt.BitsPerSample = BPS;
    Fmt.NumChannels = NC;
    Fmt.BytesPerContainer = (BPC != 0)? BPC : (Fmt.BitsPerSample + 7) / 8;

    Fmt.ChannelConfig = (CC != UINT (-1))? CC : GetStdChannelMask (Fmt.NumChannels);

    Debug (7, L"WF initialized: %s", WfxText (*this));

    Res = S_OK;

  } OnlyOnce;

  Assert (Res != E_I_PROCESSING_NOT_FINISHED);

  return Res;

}





UINT WaveFormat::GetStdChannelMask (UINT const NumChans) {

  Assert (NumChans >= NumChannelsMin && NumChans <= NumChannelsMax);

  static const UINT MaskTable [] = {

    KSAUDIO_SPEAKER_MONO,       // 1
    KSAUDIO_SPEAKER_STEREO,     // 2
    KSAUDIO_SPEAKER_DIRECTOUT,  // 3
    KSAUDIO_SPEAKER_QUAD,       // 4
    KSAUDIO_SPEAKER_DIRECTOUT,  // 5
    KSAUDIO_SPEAKER_5POINT1_SURROUND, // 6
    KSAUDIO_SPEAKER_DIRECTOUT,  // 7
    KSAUDIO_SPEAKER_7POINT1_SURROUND, // 8

  };

  return (NumChans <= NumArrayElems (MaskTable))? (
    MaskTable [NumChans - 1]
  ) : (
    KSAUDIO_SPEAKER_DIRECTOUT
  );

}





HRESULT WaveFormat::FromWfx (WAVEFORMATEX const & Wfx) {

  HRESULT Res = E_I_PROCESSING_NOT_FINISHED;

  do {

    if (Wfx.wFormatTag != WAVE_FORMAT_PCM && Wfx.wFormatTag == WAVE_FORMAT_EXTENSIBLE) {

      Res = HRESULT_FROM_WIN32 (ERROR_BAD_FORMAT);

      Debug (9, L"Invalid format tag %u", Wfx.wFormatTag);

      break;

    }

    WAVEFORMATEXTENSIBLE const & FmtExt = (WAVEFORMATEXTENSIBLE const &) (Wfx);

    bool const ExtFmt = (
      Wfx.wFormatTag != WAVE_FORMAT_PCM
      || Wfx.wBitsPerSample != 8 && Wfx.wBitsPerSample != 16
      || Wfx.nChannels != 1 && Wfx.nChannels != 2
    );

    // Never touch cbSize and other extended fields if format tag is PCM

    UINT const BPS = (ExtFmt && Wfx.wFormatTag != WAVE_FORMAT_PCM && Wfx.cbSize >= sizeof (WAVEFORMATEXTENSIBLE))? (
      FmtExt.Samples.wValidBitsPerSample
    ) : (
      Wfx.wBitsPerSample
    );

    Res = Init (Wfx.nSamplesPerSec, BPS, Wfx.nChannels, UINT (Wfx.nBlockAlign / Wfx.nChannels));

    if (FAILED (Res)) {

      break;

    }

    Assert (SUCCEEDED (Res));

  } OnlyOnce;

  Assert (Res != E_I_PROCESSING_NOT_FINISHED);

  return Res;

}





void WaveFormat::ToWfx (WAVEFORMATEXTENSIBLE & Wfx) const {

  Wfx.Format.nSamplesPerSec = GetSR ();
  Wfx.Format.nChannels = static_cast <WORD> (GetNC ());
  Wfx.Format.wBitsPerSample = static_cast <WORD> (GetBPC () * 8);
  Wfx.Format.nBlockAlign = static_cast <WORD> (GetBPC () * GetNC ());
  Wfx.Format.nAvgBytesPerSec = GetSR () * Wfx.Format.nBlockAlign;

  if (
    (GetBPS () == 8 || GetBPS () == 16)
    && (GetNC () == 1 || GetNC () == 2)
    && GetBPS () == GetBPC () * 8
  ) {

    Wfx.Format.wFormatTag = WAVE_FORMAT_PCM;
    Wfx.Format.cbSize = 0;

  } else {

    Wfx.Format.wFormatTag = WAVE_FORMAT_EXTENSIBLE;

    Wfx.SubFormat = KSDATAFORMAT_SUBTYPE_PCM; // Use INIT_WAVEFORMATEX_GUID for non-PCM
    Wfx.Format.cbSize = sizeof (Wfx) - sizeof (Wfx.Format);
    Wfx.Samples.wValidBitsPerSample = static_cast <WORD> (GetBPS ());
    Wfx.dwChannelMask = UINT (-1) >> (32 - GetNC ());

  }

}





void WaveFormat::ToKdf (KSDATAFORMAT_WAVEFORMATEXTENSIBLE & Kdf) const {

  Assert (&Kdf != nullptr);

  Debug (6, L"Creating KSDATAFORMAT_WAVEFORMATEXTENSIBLE");

  ToWfx (Kdf.WaveFormatExt);

  Kdf.DataFormat.MajorFormat = KSDATAFORMAT_TYPE_AUDIO;

  if (
    Kdf.WaveFormatExt.Format.wFormatTag != WAVE_FORMAT_PCM 
    && Kdf.WaveFormatExt.Format.cbSize >= sizeof (Kdf.WaveFormatExt) - sizeof (Kdf.WaveFormatExt.Format)
  ) {

    Kdf.DataFormat.SubFormat = Kdf.WaveFormatExt.SubFormat;

    Debug (6, L"KDF - copied subformat %s", GuidText (Kdf.DataFormat.SubFormat));

  } else {

    Assert (Kdf.WaveFormatExt.Format.cbSize == 0);

    Kdf.DataFormat.SubFormat = KSDATAFORMAT_SUBTYPE_PCM;

    Debug (6, L"KDF - subformat set to %s", GuidText (Kdf.DataFormat.SubFormat));

  }

  Kdf.DataFormat.Specifier = KSDATAFORMAT_SPECIFIER_WAVEFORMATEX;
  Kdf.DataFormat.Flags = 0;
  Kdf.DataFormat.SampleSize = 0;
  Kdf.DataFormat.Reserved = 0;

  Kdf.DataFormat.FormatSize = (
    sizeof (Kdf.DataFormat)
    + sizeof (Kdf.WaveFormatExt.Format)
    + Kdf.WaveFormatExt.Format.cbSize
  );

}





PCTSTR WaveFormat::ToStr (PTSTR const Buffer, UINT const MaxLen) const {

  _snwprintf (
    Buffer, MaxLen,
    L"%u/%u(%u)/%u(%x)",
    GetSR (), GetBPS (), GetBPC (), GetNC (), GetCC ()
  );

  return Buffer;

}





} // namespace
